"""Contains tests for the Google Cloud (gcp) storage module.
"""
import logging
import os
import sys
try:
    from ..test import AriesTest
    from ..storage import StorageFolder, StorageFile
except:
    aries_parent = os.path.join(os.path.dirname(__file__), "..", "..")
    if aries_parent not in sys.path:
        sys.path.append(aries_parent)
    from Aries.test import AriesTest
    from Aries.storage import StorageFolder, StorageFile
logger = logging.getLogger(__name__)


class TestStorageCrossPlatform(AriesTest):
    """Contains test cases for HTTP and FTP.
    """

    def test_md5(self):
        """
        """
        test_file = os.path.join(os.path.dirname(__file__), "fixtures", "links.md")
        # Local
        local_file = StorageFile(test_file)
        local_md5 = local_file.md5_hex
        self.assertIsNotNone(local_md5)

        # GCP
        gs_path = "gs://aries_test/links.md"
        local_file.copy(gs_path)
        gs_file = StorageFile(gs_path)
        self.assertEqual(local_md5, gs_file.md5_hex)
        gs_file.delete()

        # AWS
        s3_path = "s3://davelab-test/links.md"
        local_file.copy(s3_path)
        s3_file = StorageFile(s3_path)
        self.assertEqual(local_md5, s3_file.md5_hex)
        s3_file.delete()



"""Contains tests for the Google Cloud (gcp) storage module.
"""
import logging
import os
import sys

aries_parent = os.path.join(os.path.dirname(__file__), "..", "..")
if aries_parent not in sys.path:
    sys.path.append(aries_parent)
from Aries.test import AriesTest
from Aries.storage import StorageFolder, StorageFile
logger = logging.getLogger(__name__)


class TestWebStorage(AriesTest):
    """Contains test cases for HTTP and FTP.
    """

    test_folder_path = os.path.join(os.path.dirname(__file__), "fixtures", "test_folder")
    test_folder = StorageFolder(test_folder_path)
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        if not cls.test_folder.exists():
            cls.test_folder.create()

    def test_http(self):
        """
        """
        # URL does not exist
        storage_obj = StorageFile("http://example.com/abc/")
        self.assertFalse(storage_obj.exists())

        # URL exists
        storage_obj = StorageFile("https://www.google.com")
        self.assertTrue(storage_obj.exists())

        # Download. Copy to local file.
        storage_obj = StorageFile("https://www.w3.org/WAI/ER/tests/xhtml/testfiles/resources/pdf/dummy.pdf")
        local_file_path = os.path.join(self.test_folder_path, "test.pdf")
        if os.path.exists(local_file_path):
            os.remove(local_file_path)
        storage_obj.copy(local_file_path)
        self.assertTrue(os.path.exists(local_file_path))
        self.assertGreater(StorageFile(local_file_path).size, 0)
        StorageFile(local_file_path).delete()



"""This module has been moved to Aries.storage package.
The imports in this files are for backward compatibility purpose.
Use Aries.storage module for new developments.
This file will be removed in the future.
"""
from ..storage.gs import GSObject, GSFolder, GSFile



import warnings
import logging
import numpy as np
import scipy.stats
from pprint import pprint
logger = logging.getLogger(__name__)


def continuous_rvs():
    """Gets a list of all continuous random variables from scipy
    
    Returns:
        list: A list of sub-classes of scipy.stats.rv_continuous
    """
    rv_list = []
    for attr in dir(scipy.stats):
        f = getattr(scipy.stats, attr)
        if not callable(f):
            continue
        if issubclass(f.__class__, scipy.stats.rv_continuous):
            rv_list.append(f)
            continue
    return rv_list


def discrete_rvs():
    """Gets a list of all discrete random variables from scipy
    
    Returns:
        list: A list of sub-classes of scipy.stats.rv_discrete
    """
    rv_list = []
    for attr in dir(scipy.stats):
        f = getattr(scipy.stats, attr)
        if not callable(f):
            continue
        if issubclass(f.__class__, scipy.stats.rv_discrete):
            rv_list.append(f)
            continue
    return rv_list


def random_variables():
    """Gets a list of all random variables from scipy
    
    Returns:
        list: A list of sub-classes of scipy.stats.rv_continuous and scipy.stats.rv_discrete
    """
    rv_list = continuous_rvs()
    rv_list.extend(discrete_rvs())
    return rv_list

def fit_test_continuous_rv():
    """A list of default continuous random variables for fit_distribution()
    
    Returns:
        list: A list of continuous random variables
    """
    names = [
        "norm",
        "lognorm",
        "logistic",
        "laplace",
        "expon",
        "uniform",
        "levy",
        "cauchy",
    ]
    return [getattr(scipy.stats, name) for name in names]


def fit_distributions(samples, rv_list=None, filter_warnings='ignore'):
    """Fits a list of samples (numbers) to a list of random variable distribution.
    The "fitness" is determined by K-S test.
    # TODO: Use other goodness of fit tests.

    The data should not be tested on all random variable distributions.
    Because some random variables can produce the same distribution,
        when the parameters are selected carefully.
    Therefore, the numbers may fit multiple distributions in the same family.
    This function has an optional argument rv_list, 
        which should be a set of random variables from different families.
    Without specifying the rv_list parameter, 
        this function will use a small set of continuous variables returned by
        fit_test_continuous_rv()
    
    Args:
        samples ([type]): [description]
        rv_list ([type], optional): A list of candidate random variables. Defaults to None.
            If rv_list is None, the list returned by fit_test_continuous_rv() will be used.
        filter_warnings (str, optional): How to handle scipy warnings. Defaults to 'ignore'.
    
    Returns:
        list: A list of dictionaries, sorted by KS test D value.
            Each dictionary contains:
            name: The name of the fitted rv
            The best fit parameters
            The KS test D value
            The KS test p value
    """
    fits = []
    best_fit_rv = None
    best_d = 1
    best_parameters = None
    if rv_list == None:
        rv_list = fit_test_continuous_rv()
    total = len(rv_list)
    logger.debug("Fitting data to %d distributions." % total)
    counter = 1
    for rv in rv_list:
        logger.debug("Fitting data to %s...(%d/%d)" % (rv.name, counter, total))
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings(filter_warnings)
                if not hasattr(rv, "fit"):
                    continue
                parameters = rv.fit(samples)
                d, p = scipy.stats.kstest(samples, lambda x: rv.cdf(x, *parameters))
                fits.append({
                    "name": rv.name,
                    "parameters": parameters,
                    "D": d,
                    "p_value": p
                })
        except:
            logger.debug("Cannot fit to %s" % rv.name)
        counter += 1
    fits = sorted(fits, key=lambda i : i.get("D"))
    return fits


import plotly.graph_objs as go
from plotly.offline import init_notebook_mode, plot, iplot


def in_ipython_notebook():
    """Checks if the program is running in a ipython/jupyter notebook.
    """
    try:
        ipython_class = get_ipython().__class__.__name__
        if 'ZMQInteractiveShell' in ipython_class:
            return True
        else:
            return False
    except NameError:
        return False


def init_notebook():
    """Initialize plotly in a jupyter notebook for each cell.
    In Jupyter notebook, plotly can be initialized for the entire notebook using
    init_notebook_mode(connected=True)
    However, in Google Colab, the initialization needs to be done in each cell.
    IPython allows user to register event callbacks for each cell.
    This cell register the initialization as the "pre_execute" event callback.
    See Also: https://ipython.readthedocs.io/en/stable/config/callbacks.html
    """
    # Import IPython here so that it is not required.
    import IPython
    def init_plotly():
        display(IPython.core.display.HTML(
            '''
            <script src="/static/components/requirejs/require.js"></script>
            <script>
            requirejs.config({
                paths: {
                base: '/static/base',
                plotly: 'https://cdn.plot.ly/plotly-latest.min.js?noext',
                },
            });
            </script>
            '''
        ))
    IPython.get_ipython().events.register('pre_execute', init_plotly)


class PlotlyFigure:
    """Represents a Plotly figure and provides helper methods for plotting.

    Many methods in this class support "Method Chaining", i.e. they return the PlotlyFigure instance itself.

    Attributes:
        __notebook_mode: Indicates if the plotly notebook mode has been initialized.

    Examples:
        PlotlyFigure().candle_stick(ts.df, "Daily Data").lines(tech_df).plot()

    """

    __notebook_mode = False

    def __init__(self, plotly_traces=None, plotly_layout=None, **kwargs):
        if plotly_traces:
            self.data = plotly_traces
        else:
            self.data = []
        self.layout = plotly_layout
        self.layout_args = kwargs

        self.title = ""
        self.title_x = ""
        self.title_y = ""

    @staticmethod
    def __is_pandas_data_frame(x):
        """Checks if x and args are pandas data frames using strings to avoid importing pandas.
        So that pandas is not required in order to use this function.

        Args:
            x: A variable.

        Returns: True if "pandas" and "DataFrame" are both in the type/class name of x

        """
        from pandas import DataFrame
        return issubclass(x.__class__, DataFrame)

    @staticmethod
    def __is_pandas_data_series(x):
        from pandas import Series
        return issubclass(x.__class__, Series)

    @property
    def figure(self):
        """Prepares the plotly figure.
        """
        if self.layout is None:
            self.layout = go.Layout(
                title=self.title,
                xaxis=dict(
                    rangeslider=dict(
                        visible=False
                    ),
                    title=self.title_x,
                    automargin=True
                ),
                yaxis=dict(
                    title=self.title_y,
                    automargin=True
                ),
                **self.layout_args
            )
        return go.Figure(data=self.data, layout=self.layout)

    def set_title(self, title, title_x="", title_y=""):
        self.title = title
        self.title_x = title_x
        self.title_y = title_y
        return self

    def to_html(self):
        return plot(self.figure, auto_open=False, output_type='div')

    def to_file(self):
        raise NotImplementedError

    def plot(self):
        """Plots the figure in a jupyter notebook
        """
        if not self.__notebook_mode:
            init_notebook_mode(connected=True)
            self.__notebook_mode = True
        iplot(self.figure)
        return self

    # The following methods support "Method Chaining"

    def candle_stick(self, df, name=None, **kwargs):
        if not self.title_x:
            self.title_x = "Time"
        if not self.title_y:
            self.title_y = "Value"

        trace = go.Candlestick(
            x=df.index,
            open=df.open,
            high=df.high,
            low=df.low,
            close=df.close,
            name=name,
            **kwargs
        )
        self.data.append(trace)

        return self

    def line(self, x, y, name=None, **kwargs):
        trace = go.Scatter(
            x=x,
            y=y,
            name=name,
            **kwargs
        )
        self.data.append(trace)
        return self

    def __parse_args(self, names, array, *args):
        for arg in args:
            if isinstance(arg, str):
                names.append(arg)
            else:
                array.append(arg)
        return names, array

    def lines(self, x, *args, **kwargs):
        """

        Args:
            x:
            *args:

        Returns:

        """
        y_list = []
        y_name = []

        if self.__is_pandas_data_frame(x):
            df = x
            x = df.index
            for c in df.columns.values:
                y_list.append(df[c])
                y_name.append(c)
        elif self.__is_pandas_data_series(x):
            series = x
            x = series.index
            y_list.append(series)
            y_name.append(series.name)
        else:
            y_name, y_list = self.__parse_args(y_name, y_list, *args)
        for i in range(len(y_list)):
            trace = go.Scatter(
                x=x,
                y=y_list[i],
                name=y_name[i] if i < len(y_name) else None,
                **kwargs
            )
            self.data.append(trace)
        return self

    def bar(self, x, y, name=None, **kwargs):
        trace = go.Bar(x=x, y=y, name=name, **kwargs)
        self.data.append(trace)
        return self

    def add_trace(self, trace_type, **kwargs):
        trace = getattr(go, trace_type)(**kwargs)
        self.data.append(trace)
        return self



"""Contains classes for manipulating Google Cloud Storage Objects.

If you are not running in Google Compute Engine or App Engine,
authentication to Google Cloud Platform is required in order to use this module.

Authentication can be done by setting the "GOOGLE_APPLICATION_CREDENTIALS" environment variable
to JSON key-file.
In command line:
    $ export GOOGLE_APPLICATION_CREDENTIALS="/path/to/keyfile.json"
In python:
    os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = "/path/to/keyfile.json"

See Also:
    https://googleapis.github.io/google-cloud-python/latest/core/auth.html
    https://googleapis.dev/python/storage/latest/
    https://github.com/googleapis/python-storage

"""
import os
import logging
import warnings
import tempfile
import base64
import binascii
from functools import wraps
from google.cloud import storage
from google.cloud.exceptions import ServerError
from ..strings import Base64String
from ..tasks import FunctionTask
from .base import StorageFolderBase
from .cloud import BucketStorageObject, CloudStoragePrefix, CloudStorageIO
logger = logging.getLogger(__name__)


def setup_credentials(env_name, to_json_file=None):
    """Configures the GOOGLE_APPLICATION_CREDENTIALS
    by saving the value of an environment variable to a JSON file.
    """
    # Use the b64 encoded content as credentials if "GOOGLE_CREDENTIALS" is set.
    credentials = os.environ.get(env_name)
    if credentials and credentials.startswith("ew"):
        if not to_json_file:
            temp_file = tempfile.NamedTemporaryFile(suffix=".json", delete=False)
            temp_file.close()
            to_json_file = temp_file.name
        Base64String(credentials).decode_to_file(to_json_file)
    # Set "GOOGLE_APPLICATION_CREDENTIALS" if json file exists.
    if os.path.exists(to_json_file):
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = to_json_file


def api_call(func=None, *args, **kwargs):
    """Makes API call and retry if there is an exception.
    This is designed to resolve the 500 Backend Error from Google.

    Args:
        func (callable): A function or method.

    Examples:
        api_call(self.bucket.get_blob, self.prefix)

    See Also: https://developers.google.com/drive/api/v3/handle-errors#resolve_a_500_error_backend_error
    """
    if not func:
        return None
        # logger.debug("Making API call: %s..." % func.__name__)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", ResourceWarning)
        warnings.simplefilter("ignore", UserWarning)
        return FunctionTask(func, *args, **kwargs).run_and_retry(
            max_retry=3,
            exceptions=ServerError,
            base_interval=60,
            retry_pattern='linear'
        )


def api_decorator(method):
    """Decorator for making API call and retry if there is an exception.
    This is designed to resolve the 500 Backend Error from Google.
    When the decorated function is called, the function call will be retry if there is an exception.

    Examples:
        @api_decorator
        def exists(self):
            return self.blob.exists
    """
    # logger.debug("Decorating %s for API call..." % method.__name__)
    @wraps(method)
    def wrapper(*method_args, **method_kwargs):
        return api_call(method, *method_args, **method_kwargs)
    return wrapper


class GSObject(BucketStorageObject):
    """The base class for Google Storage Object.
    """
    MAX_BATCH_SIZE = 900

    @property
    def blob(self):
        """Gets or initialize a Google Cloud Storage Blob.

        Returns: A Google Cloud Storage Blob object.

        This does not check whether the object exists.
        Use blob.exists() to determine whether or not the blob exists.

        """
        if self._blob is None:
            # logger.debug("Getting blob: %s" % self.uri)
            file_blob = api_call(self.bucket.get_blob, self.prefix)
            if file_blob is None:
                # The following will not make an HTTP request.
                # It simply instantiates a blob object owned by this bucket.
                # See https://googleapis.github.io/google-cloud-python/latest/storage/buckets.html
                # #google.cloud.storage.bucket.Bucket.blob
                file_blob = self.bucket.blob(self.prefix)
            self._blob = file_blob
        return self._blob

    @api_decorator
    def init_client(self):
        return storage.Client()

    @api_decorator
    def init_bucket(self):
        return self.client.get_bucket(self.bucket_name)

    @property
    def gs_path(self):
        return self.uri

    @api_decorator
    def exists(self):
        """Determines if there is an actual blob corresponds to this object.
        """
        return self.blob.exists()

    @api_decorator
    def create(self):
        """Creates an empty blob, if the blob does not exist.

        Returns:
            Blob: The Google Cloud Storage blob.
        """
        blob = storage.Blob(self.prefix, self.bucket)
        if not blob.exists():
            blob.upload_from_string("")
        return blob

    def delete(self):
        self.delete_blob(self.blob)

    def copy(self, to):
        self.copy_blob(self.blob, to)

    def copy_blob(self, blob, to):
        """Copies a blob object in the bucket to a new location.

        Args:
            blob: A Google Cloud Storage Blob object in the bucket.
            to: URI of the new blob (gs://...).

        Returns: True if the blob is copied. Otherwise False.

        """
        destination = GSObject(to)
        new_name = str(blob.name).replace(self.prefix, destination.prefix, 1)
        if new_name != str(blob.name) or self.bucket_name != destination.bucket_name:
            self.bucket.copy_blob(blob, destination.bucket, new_name)
            return True
        return False

    @staticmethod
    def delete_blob(blob):
        blob.delete()


class GSPrefix(CloudStoragePrefix, GSObject):
    # @api_decorator
    def batch_request(self, blobs, method, *args, **kwargs):
        """Sends a batch request to run method of a batch of blobs.
        The "method" will be applied to each blob in blobs like method(blob, *args, **kwargs)

        Args:
            blobs: A list of blobs, to be processed in a SINGLE batch.
            method: The method for processing each blob.
            *args: Additional arguments for method.
            **kwargs: Keyword arguments for method.

        Returns:

        """
        if not blobs:
            return 0
        counter = 0
        with self.client.batch():
            for blob in blobs:
                method(blob, *args, **kwargs)
                counter += 1
        return counter

    # @api_decorator
    def batch_operation(self, method, *args, **kwargs):
        blobs = self.blobs()
        batch = []
        counter = 0
        for blob in blobs:
            batch.append(blob)
            if len(batch) > self.MAX_BATCH_SIZE:
                counter += self.batch_request(batch, method, *args, **kwargs)
                batch = []
        if batch:
            counter += self.batch_request(batch, method, *args, **kwargs)
        return counter

    @api_decorator
    def blobs(self, delimiter=None):
        """Gets the blobs in the bucket having the prefix.

        The returning list will contain object in the folder and all sub-folders

        Args:
            delimiter: Use this to emulate hierarchy.
            If delimiter is None, the returning list will contain objects in the folder and in all sub-directories.
            Set delimiter to "/" to eliminate files in sub-directories.

        Returns: A list of GCS blobs.

        See Also: https://googleapis.github.io/google-cloud-python/latest/storage/blobs.html

        """
        return list(self.bucket.list_blobs(prefix=self.prefix, delimiter=delimiter))

    @property
    def uri_list(self):
        """Gets all file URIs with the prefix
        """
        return [
            "gs://%s/%s" % (self.bucket_name, b.name)
            for b in self.blobs()
            if not b.name.endswith("/")
        ]

    @property
    def files(self):
        from .io import StorageFile
        storage_files = []
        for b in self.blobs("/"):
            if b.name.endswith("/"):
                continue
            storage_file = StorageFile("gs://%s/%s" % (self.bucket_name, b.name))
            storage_file.raw_io._blob = b
            storage_files.append(storage_file)
        return storage_files

    @property
    def folders(self):
        return self.list_folders

    @api_decorator
    def list_folders(self):
        from .io import StorageFolder
        iterator = self.bucket.list_blobs(prefix=self.prefix, delimiter='/')
        list(iterator)
        return [
            StorageFolder("gs://%s/%s" % (self.bucket_name, p))
            for p in iterator.prefixes
        ]

    def exists(self):
        return True if self.blob.exists() or self.objects else False

    @api_decorator
    def delete(self):
        """Deletes all objects with the same prefix."""
        counter = self.batch_operation(self.delete_blob)
        logger.debug("%d files deleted." % counter)
        return counter

    @api_decorator
    def copy(self, to, contents_only=False):
        """Copies folder/file in a Google Cloud storage directory to another one.

        Args:
            to (str): Destination Google Cloud Storage path.
            contents_only: Copies only the content of the folder. This applies only if the GSObject is a folder.
                Defaults to False, i.e. a folder (with the same name as this folder)
                will be created at the destination to contain the files.

        Returns: The number of files copied.

        Warnings:
            When the URI of GSObject ends with "/", i.e. it is a folder,
                use "contents_only" to indicate if a new folder should be created to contain all files copied.
            When the GSObject is a file or a set of filtered files with the same prefix:
                If to ends with slash ("/"), all files will be copied under the "to" folder.
                    folders partially in the prefix will be kept.
                If to does NOT end with slash, the prefix of all files will simply be replaced with the prefix in "to".
                See the following examples for more details.

        Example:
            Either
                GSFolder("gs://bucket_a/alpha/beta/").copy("gs://bucket_b/x/y")
            or
                GSFolder("gs://bucket_a/alpha/beta/").copy("gs://bucket_b/x/y/")
            will copy the following files:
                gs://bucket_a/alpha/beta/gamma/example.txt
                gs://bucket_a/alpha/beta/example.txt
            to
                gs://bucket_b/x/y/beta/gamma/example.txt
                gs://bucket_b/x/y/beta/example.txt

            Either
                GSFolder("gs://bucket_a/alpha/beta/").copy("gs://bucket_b/x/y", contents_only=True)
            or
                GSFolder("gs://bucket_a/alpha/beta/").copy("gs://bucket_b/x/y/", contents_only=True)
            will copy the following files:
                gs://bucket_a/alpha/beta/gamma/example.txt
                gs://bucket_a/alpha/beta/example.txt
            to
                gs://bucket_b/x/y/gamma/example.txt
                gs://bucket_b/x/y/example.txt

            Also
                GSFolder("gs://bucket_a/alpha/be").copy("gs://bucket_b/x/y/")
            will copy the following files:
                gs://bucket_a/alpha/beta/gamma/example.txt
                gs://bucket_a/alpha/beta/example.txt
            to
                gs://bucket_b/x/y/beta/gamma/example.txt
                gs://bucket_b/x/y/beta/example.txt

            However
                GSFolder("gs://bucket_a/alpha/be").copy("gs://bucket_b/x/y")
            will copy the following files:
                gs://bucket_a/alpha/beta/gamma/example.txt
                gs://bucket_a/alpha/beta/example.txt
            to
                gs://bucket_b/x/yta/gamma/example.txt
                gs://bucket_b/x/yta/example.txt

        """
        # Check if the destination is a bucket root.
        # Prefix will be empty if destination is bucket root.
        # Always append "/" to bucket root.
        if not GSObject(to).prefix and not to.endswith("/"):
            to += "/"

        if self.prefix.endswith("/"):
            # The source is a folder if its prefix ends with "/"
            if contents_only:
                to += "/"
            else:
                # Copy the contents into a folder with the same name.
                to = os.path.join(to, self.name) + "/"
        else:
            # Otherwise, it can be a file or an object or a set of filtered objects or a folder.
            if to.endswith("/"):
                # If the destination ends with "/",
                # copy all objects under the destination
                to += self.name
            else:
                # If the destination does not end with "/",
                # simply replace the prefix.
                pass
        # logger.debug("Copying files to %s" % to)
        source_files = self.blobs()
        if not source_files:
            logger.debug("No files in %s" % self.uri)
            return 0
        counter = self.batch_operation(self.copy_blob, to)
        logger.debug("%d files copied." % counter)
        return counter


class GSFolder(GSPrefix, StorageFolderBase):
    """Represents a Google Cloud Storage Folder

    Method Resolution Order: GSFolder, GSObject, StorageFolder, StorageObject
    """

    def __init__(self, uri):
        """Initializes a Google Cloud Storage Directory.

        Args:
            uri: The path of the object, e.g. "gs://bucket_name/path/to/dir/".

        """
        # super() will call the __init__() of StorageObject, StorageFolder and GSObject
        GSObject.__init__(self, uri)
        StorageFolderBase.__init__(self, uri)

        # Make sure prefix ends with "/", otherwise it is not a "folder"
        if self.prefix and not self.prefix.endswith("/"):
            self.prefix += "/"

    def exists(self):
        return True if self.blob.exists() or self.file_paths or self.folder_paths else False

    @property
    def folder_paths(self):
        """Folders(Directories) in the directory.
        """
        return self.__folders_paths()

    @api_decorator
    def __folders_paths(self):
        iterator = self.bucket.list_blobs(prefix=self.prefix, delimiter='/')
        list(iterator)
        return [
            "gs://%s/%s" % (self.bucket_name, p)
            for p in iterator.prefixes
        ]

    @property
    def file_paths(self):
        """Files in the directory
        """
        paths = self.__file_paths()
        return paths

    def __file_paths(self):
        return [
            "gs://%s/%s" % (self.bucket_name, b.name)
            for b in self.blobs("/")
            if not b.name.endswith("/")
        ]

    @api_decorator
    def filter_files(self, prefix):
        return [
            GSFile("gs://%s/%s" % (self.bucket_name, b.name))
            for b in self.bucket.list_blobs(prefix=os.path.join(self.prefix, prefix), delimiter='/')
            if not b.name.endswith("/")
        ]


class GSFile(GSObject, CloudStorageIO):
    def __init__(self, uri):
        """Represents a file on Google Cloud Storage as a file-like object implementing the IOBase interface.

        Args:
            uri:

        GSFile allows seek and read without opening the file.
        However, position/offset will be reset when open() is called.
        The context manager calls open() when enter.
        """
        GSObject.__init__(self, uri)
        CloudStorageIO.__init__(self, uri)

    @property
    def updated_time(self):
        return self.blob.updated

    @property
    def md5_hex(self):
        return binascii.hexlify(base64.urlsafe_b64decode(self.blob.md5_hash)).decode()

    def get_size(self):
        return self.blob.size

    def read_bytes(self, start, end):
        return api_call(self.blob.download_as_string, start=start, end=end)

    def download(self, to_file_obj):
        api_call(self.blob.download_to_file, to_file_obj)
        return to_file_obj

    def upload(self, from_file_obj):
        api_call(self.blob.upload_from_file, from_file_obj)



"""Contains tests for the strings module.
"""
import datetime
import logging

import os
import sys
aries_parent = os.path.join(os.path.dirname(__file__), "..", "..")
if aries_parent not in sys.path:
    sys.path.append(aries_parent)
from Aries.test import AriesTest
from Aries.strings import AString, FileName

logger = logging.getLogger(__name__)


class TestAString(AriesTest):
    def test_str_methods(self):
        """Tests calling methods of python str with AString
        """
        # Tests cases for methods returning a string
        test_cases = [
            # (input_value, method_to_call, expected_output)
            ("test string", "title", "Test String"),
            ("test String", "lower", "test string"),
        ]
        for t in test_cases:
            value = t[0]
            attr = t[1]
            expected = t[2]
            x = AString(value)
            actual = getattr(x, attr)()
            self.assertEqual(type(actual), AString)
            self.assertEqual(str(actual), expected, "Test Failed. Attribute: %s" % attr)

        # Test split() method, which returns a list
        arr = AString("test string").split(" ")
        self.assertTrue(isinstance(arr, list), "split() should return a list.")
        self.assertEqual(len(arr), 2, "split() should return a list of 2 elements.")
        self.assertEqual(type(arr[0]), AString)
        self.assertEqual(type(arr[1]), AString)
        self.assertEqual(arr[0], "test", "split() Error.")
        self.assertEqual(arr[1], "string", "split() Error.")

        # Test the partition() method, which returns a 3-tuple
        arr = AString("test partition string").partition("partition")
        self.assertTrue(isinstance(arr, tuple), "partition() should return a tuple.")
        self.assertEqual(len(arr), 3, "partition() should return a list of 3 elements.")
        self.assertEqual(arr[0], "test ", "partition() Error.")
        self.assertEqual(arr[1], "partition", "partition() Error.")
        self.assertEqual(arr[2], " string", "partition() Error.")

        # Test endswith() method
        self.assertTrue(AString("test string").endswith("ing"), "endswith() Error.")

    def test_append_strings(self):
        """Tests appending strings
        """
        test_string = "test"
        a_string = AString(test_string)
        # Append date
        output = a_string.append_today()
        today = datetime.datetime.today()
        self.assertEqual(output, "%s_%s%s%s" % (
            test_string,
            str(today.year),
            str(today.month).zfill(2),
            str(today.day).zfill(2)
        ))

    def test_prepend_strings(self):
        """Tests prepending strings
        """
        test_string = "test"
        a_string = AString(test_string)
        # Prepend a list of strings
        prepend_list = ["a", "bc", "def"]
        output = a_string.prepend(prepend_list)
        self.assertEqual(str(output), "a_bc_def_test")
        # Prepend a single string
        self.assertEqual(a_string.prepend("abc"), "abc_test")

    def test_remove_non_alphanumeric(self):
        """Tests removing characters.
        """
        test_string = "!te\nst.123"
        self.assertEqual(
            AString(test_string).remove_non_alphanumeric(), 
            "test123"
        )

    def test_remove_escape_sequence(self):
        """Tests removing ANSI escape sequence.
        """
        test_string = "test\r\n\x1b[00m\x1b[01;31mHello.World\x1b[00m\r\n\x1b[01;31m"
        self.assertEqual(
            AString(test_string).remove_escape_sequence(), 
            "test\r\nHello.World\r\n"
        )

    def test_remove_non_ascii(self):
        """Tests removing non ASCII characters.
        """
        test_string = "!!te‘’st\n"
        self.assertEqual(
            AString(test_string).remove_non_ascii(), 
            "!!test\n"
        )

    def test_equality(self):
        """Tests equality operator.
        """
        a = AString("equal")
        b = AString("equal")
        self.assertEqual(a, b)
        self.assertEqual(AString(""), AString(None))


class TestFileName(AriesTest):

    @staticmethod
    def run_test(assert_method):
        test_base = "abc_def"
        test_extension = ".txt"
        assert_method(test_base, "")
        assert_method(test_base, test_extension)

    def test_properties(self):
        """Tests properties of the FileName class by initializing a FileName with extension.
        """
        self.run_test(self.assert_properties)

    def assert_properties(self, basename, extension):
        # Construct the test filename
        test_name = basename + extension
        # Make sure the test_name is a python string
        self.assertTrue(isinstance(test_name, str))
        filename = FileName(test_name)

        # Test converting FileName to string
        self.assertEqual(type(filename), FileName)
        self.assertEqual(str(filename), test_name)
        self.assertEqual(filename.to_string(), test_name)
        # Test basename
        self.assertEqual(str(filename.basename), basename)
        self.assertEqual(str(filename.name_without_extension), basename)
        # Test extension
        self.assertEqual(str(filename.extension), extension)

    def test_methods(self):
        """Tests the methods returning a FileName instance.
        """
        self.run_test(self.assert_methods)

    def assert_methods(self, basename, extension):
        # Construct the test filename
        test_name = basename + extension
        # Make sure the test_name is a python string
        self.assertTrue(isinstance(test_name, str))
        filename = FileName(test_name)

        # Test str method: title()
        title = filename.title()
        self.assertEqual(type(title), FileName)
        self.assertEqual(str(title), basename.title() + extension)
        # Test str method: upper()
        self.assertEqual(str(filename.upper()), basename.upper() + extension)

        # Test AString method: append_random_letters()
        n = 5
        filename_with_random = filename.title().append_random_letters(n)
        # Append adds an underscore between the random letters and the original filename.
        self.assertEqual(len(filename_with_random), len(test_name) + n + 1, "Filename: %s" % filename_with_random)
        self.assertIn(title.basename, str(filename_with_random))
        self.assertIn(title.extension, str(filename_with_random))

    def test_split(self):
        """Tests the split method, which returns a list.
        """
        self.run_test(self.assert_split)

    def assert_split(self, basename, extension):
        # Construct the test filename
        test_name = basename + extension
        # Make sure the test_name is a python string
        self.assertTrue(isinstance(test_name, str))
        filename = FileName(test_name)

        # Test str method: split()
        filename_splits = filename.split("_")
        self.assertEqual(len(filename_splits), 2)
        str_splits = basename.split("_")
        self.assertEqual(filename_splits[0], str_splits[0])
        self.assertEqual(filename_splits[1], str_splits[1])

    def test_operators(self):
        """Tests operators
        """
        self.run_test(self.assert_operators)

    def assert_operators(self, basename, extension):
        # Construct the test filename
        test_name = basename + extension
        # Make sure the test_name is a python string
        self.assertTrue(isinstance(test_name, str))
        filename = FileName(test_name)

        # Test "+" operator
        addition = "12345"
        new_filename = filename + addition
        self.assertEqual(str(new_filename), basename + extension + addition)

        # Test "*" operator
        multiplier = 2
        new_filename = filename * multiplier
        self.assertEqual(str(new_filename), (basename + extension) * 2)

        # Test slice
        new_filename = filename[2:4]
        self.assertEqual(str(new_filename), basename[2:4])

        # Test membership
        member = filename[2:4]
        self.assertTrue(member in filename)



"""provides classes and methods for logging and capturing outputs from the code, 
like a function or a method. 
"""
import io
import json
import logging
import pprint
import sys
import threading
import traceback
import uuid
import copy
logger = logging.getLogger(__name__)


class OutputWriter(io.StringIO):
    """Represents a writer for a list of file-like objects (listeners).

    An instance of this class must be initialized with a list of file-like objects (listeners).
    Any output sending to the instance will result in writing into all listeners.

    Although this class accepts any file-like object as listener, only StringIO() is tested.
    This class does not handle the opening and closing of a file.

    """
    def __init__(self, listeners):
        """Initialize a writer with a list of file-like objects (listeners).

        Args:
            listeners (list): A list of file-like objects.
        """
        self.listeners = listeners
        super(OutputWriter, self).__init__()

    def write(self, *args, **kwargs):
        """Writes the output to the listeners.
        """
        for listener in self.listeners:
            listener.write(*args, **kwargs)


class CaptureOutput:
    """Represents an object capturing the standard outputs and standard errors.

    In Python 3.5 and up, redirecting stdout and stderr can be done by using:
        from contextlib import redirect_stdout, redirect_stderr

    Attributes:
        std_out (str): Captured standard outputs.
        std_err (str): Captured standard errors.
        log_out (str): Captured log messages.
        exc_out (str): Captured exception outputs.
        returns: This is not used directly. It can be used to store the return value of a function/method.
        log_handler (ThreadLogHandler): The log handler object for capturing the log messages.

    Class Attributes:
        sys_out: The system stdout before using any instance of this class.
            When the last instance of this class exits, the sys.stdout will be reset to CaptureOutput.sys_out.
        sys_err: The system stderr before using any instance of this class.
            When the last instance of this class exits, the sys.stderr will be reset to CaptureOutput.sys_err.
        out_listeners: A dictionary of file-like objects expecting outputs from stdout.
        err_listeners: A dictionary of file-like objects expecting outputs from stderr.

    Examples:
        with CaptureOutput() as out:
            do_something()

        standard_output = out.std_out
        standard_error = out.std_err
        log_messages = out.log_out

    Multi-Threading:
        When using this class, stdout/stderr from all threads in the same process will be captured.
        To capture the stdout/stderr of a particular thread, run the thread in an independent process.
        Only the logs of the current thread will be captured.

    Warnings:
        Using this class will set the level of root logger to DEBUG.
        sys.stdout and sys.stderr should not be modified when using this class.

    """

    sys_out = None
    sys_err = None

    out_listeners = {}
    err_listeners = {}

    def __init__(self, suppress_exception=False, log_level=logging.DEBUG, filters=None):
        """Initializes log handler and attributes to store the outputs.
        """
        self.uuid = uuid.uuid4()
        self.suppress_exception = suppress_exception

        self.log_handler = ThreadLogHandler(threading.current_thread().ident)
        self.log_handler.setLevel(log_level)
        if filters:
            if not isinstance(filters, list):
                filters = [filters]
            for log_filter in filters:
                self.log_handler.addFilter(log_filter)

        self.std_out = ""
        self.std_err = ""

        self.log_out = ""
        self.logs = []
        self.exc_out = ""
        self.returns = None

    @staticmethod
    def config_sys_outputs():
        """Configures sys.stdout and sys.stderr.
            If there are listeners, sys.stdout/sys.stderr will be redirect to an instance of OutputWriter,
            which will write the outputs to all listeners.
            If there is no listener, sys.stdout/sys.stderr will be restored to the values saved before.

        """
        if CaptureOutput.out_listeners:
            out_listener_list = [l for l in CaptureOutput.out_listeners.values()]
            out_listener_list.append(CaptureOutput.sys_out)
            sys.stdout = OutputWriter(out_listener_list)
        else:
            sys.stdout = CaptureOutput.sys_out
            CaptureOutput.sys_out = None

        if CaptureOutput.err_listeners:
            err_listener_list = [l for l in CaptureOutput.err_listeners.values()]
            err_listener_list.append(CaptureOutput.sys_err)
            sys.stderr = OutputWriter(err_listener_list)
        else:
            sys.stderr = CaptureOutput.sys_err
            CaptureOutput.sys_err = None

    def __enter__(self):
        """Configures sys.stdout and sys.stderr, and attaches the log handler to root logger.

        Returns: A CaptureOutput object (self).

        """
        # Save the sys.stdout and sys.stderr before the first instance of this class start capturing outputs.
        if CaptureOutput.sys_out is None:
            CaptureOutput.sys_out = sys.stdout
        if CaptureOutput.sys_err is None:
            CaptureOutput.sys_err = sys.stderr

        # Update listeners and re-config output writer.
        CaptureOutput.out_listeners[self.uuid] = io.StringIO()
        CaptureOutput.err_listeners[self.uuid] = io.StringIO()
        self.config_sys_outputs()

        # Modify root logger level and add log handler.
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG)
        root_logger.addHandler(self.log_handler)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Saves the outputs, configures sys.stdout and sys.stderr, and removes log handler.

        Returns: True if self.suppress_exception is True, otherwise False.

        """
        # Capture exceptions, if any
        if exc_type:
            self.exc_out = traceback.format_exc()

        # Removes log handler
        root_logger = logging.getLogger()
        root_logger.removeHandler(self.log_handler)
        self.log_out = "\n".join(self.log_handler.logs)
        self.logs = self.log_handler.logs

        # Update listeners and re-config output writer.
        self.std_out = CaptureOutput.out_listeners.pop(self.uuid).getvalue()
        self.std_err = CaptureOutput.err_listeners.pop(self.uuid).getvalue()
        self.config_sys_outputs()

        # Exception will be suppressed if returning True
        if self.suppress_exception:
            return True
        return False


class MessageFormatter(logging.Formatter):
    """Logging Formatter for pretty printing dictionary and list log message.
    By default, this formatter will show time, line number and module in addition to the log message.
    Bytes messages will also be decoded to string.

    See Also: https://docs.python.org/3/library/logging.html#formatter-objects
    """
    # Default message and date formats
    message_format = '%(asctime)s | %(levelname)-8s | %(lineno)4d@%(module)-15s | %(message)s'
    date_format = '%Y-%m-%d %H:%M:%S'
    
    def __init__(self, fmt=None, datefmt=None, style="%", encoding='utf-8'):
        """Initializes a new instance of formatter
        
        Args:
            fmt (str, optional): Log message format. 
                "MessageFormatter.message_format" attribute will be used if fmt is None.
            datefmt (str, optional): Date format. Defaults to '%Y-%m-%d %H:%M:%S'.
            style (str, optional): formatting styles. Defaults to "%".
                See: https://docs.python.org/3/howto/logging-cookbook.html#formatting-styles
            encoding (str, optional): The encoding for decoding bytes messages. Defaults to 'utf-8'.
        """
        self.encoding = encoding
        if fmt is None:
            fmt = self.message_format
        if datefmt is None:
            datefmt = self.date_format
        super().__init__(fmt, datefmt, style)

    def format(self, record):
        """Formats the log record.

        See https://docs.python.org/3/library/logging.html#logging.Formatter.format
        """
        message = record.msg
        # Decode if the message is bytes
        if self.encoding and isinstance(message, bytes):
            message = message.decode(self.encoding)
        # Try to print dict or list objects as json format with indent.
        # Use pprint if json.dumps() does not work.
        try:
            # if isinstance(message, str):
            #     message = ast.literal_eval(message)
            if isinstance(message, dict) or isinstance(message, list):
                message = json.dumps(message, sort_keys=True, indent=4)
        except Exception:
            message = pprint.pformat(message)
        # Start the message in a new line if the message contains multiple lines.
        # This will print the dict and list starting from a new line.
        if isinstance(message, str) and "\n" in message:
            record.msg = "\n" + message
        else:
            record.msg = message

        message = super(MessageFormatter, self).format(record)
        return message


class StreamHandler(logging.StreamHandler):
    """Stream Handler with customized formats to output module name and line number.
    Logs are sent to sys.stdout
    """

    def __init__(self, stream=sys.stdout, formatter=None):
        """Initialize the handler to send logging to standard output.

        Args:
            stream: The stream to which the outputs are sent.

        """
        super().__init__(stream)
        if isinstance(formatter, str):
            self.setFormatter(MessageFormatter(formatter))
        elif formatter:
            self.setFormatter(formatter)
        else:
            self.setFormatter(MessageFormatter())


class ThreadLogHandler(logging.Handler):
    """Captures the logs of a particular thread.

    Attributes:
        thread_id: The ID of the thread of which the logs are being captured.
        logs (list): A list of formatted log messages.

    Examples:
        log_handler = ThreadLogHandler(threading.current_thread().ident)
        logger = logging.getLogger(__name__)
        logger.addHandler(log_handler)

    See Also:
        https://docs.python.org/3.5/library/logging.html#handler-objects
        https://docs.python.org/3.5/library/logging.html#logrecord-attributes
        https://github.com/python/cpython/blob/master/Lib/logging/__init__.py

    """
    def __init__(self, thread_id, formatter=None):
        """Initialize the log handler for a particular thread.

        Args:
            thread_id: The ID of the thread.
        """
        super(ThreadLogHandler, self).__init__()
        if formatter is None:
            formatter = MessageFormatter()
        elif isinstance(formatter, str):
            formatter = MessageFormatter(formatter)
        self.setFormatter(formatter)
        self.thread_id = thread_id
        self.logs = []

    def handle(self, record):
        """Determine whether to emit base on the thread ID.
        """
        if record.thread == self.thread_id:
            return super().handle(record)
        return False

    def emit(self, record):
        """Formats and saves the log message.
        """
        try:
            message = self.format(record)
            self.logs.append(message)
        except Exception:
            self.handleError(record)


class PackageLogFilter(logging.Filter):
    """Logging filter to keep logs generated from packages within a certain location
    
    """
    def __init__(self, package_root=None, packages=None):
        """Initialize the filter with a folder path containing the packages.
        This filter keeps only the logs from the packages within the folder.
        
        Args:
            package_root (str): Full path of a folder

        """
        if package_root:
            self.packages = self.get_packages(package_root)
        else:
            self.packages = []
        if isinstance(packages, list):
            self.packages.extend(packages)
        else:
            self.packages.append(packages)

        logger.info("Filtering logs except packages: %s" % self.packages)
        super().__init__()

    def filter(self, record):
        # Do not filter logs that are not in a package.
        if "." not in record.name:
            return True
        logger_name = record.name.split(".", 1)[0]
        if logger_name in self.packages:
            return True
        return False

    @staticmethod
    def get_packages(folder_path):
        from .storage import StorageFolder
        project_packages = []
        project_folder = StorageFolder(folder_path)
        sub_folders = project_folder.folders
        for sub_folder in sub_folders:
            if "__init__.py" in sub_folder.file_names:
                project_packages.append(sub_folder.name)
        return project_packages


class LoggingConfig:
    """A helper class for configuring python logging.
    """
    def __init__(self, name="", level=logging.DEBUG, formatter=None, filters=None):
        """Initializes a logging config
        
        Args:
            name (str, optional): [description]. Defaults to "".
            level ([type], optional): [description]. Defaults to logging.DEBUG.
            formatter ([type], optional): [description]. Defaults to None.
        """
        self.name = name
        self.level = level
        self.existing_level = None
        self.stream_handler = StreamHandler(formatter=formatter)
        self.log_filters = filters
        if not self.log_filters:
            self.log_filters = []

    def enable(self):
        """Adds a stream_handler to format the logging outputs.
        """
        named_logger = logging.getLogger(self.name)
        self.existing_level = named_logger.getEffectiveLevel()
        named_logger.setLevel(self.level)
        # Add log filters
        for log_filter in self.log_filters:
            self.stream_handler.addFilter(log_filter)
        named_logger.addHandler(self.stream_handler)
        
        return self

    def disable(self):
        """Removes the stream_handler added by enable()
        """
        named_logger = logging.getLogger(self.name)
        named_logger.removeHandler(self.stream_handler)
        for log_filter in self.log_filters:
            named_logger.removeFilter(log_filter)
        named_logger.setLevel(self.existing_level)

    def __enter__(self):
        """Adds a stream_handler to format the logging outputs.
        """
        return self.enable()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Removes the stream_handler added by enable()
        """
        self.disable()
        return False

    @staticmethod
    def decorate(func, name="", level=logging.DEBUG, formatter=None):
        def wrapper(*args, **kwargs):
            with LoggingConfig(name, level, formatter):
                values = func(*args, **kwargs)
            return values
        return wrapper


class LoggingConfigDict:
    default_config_dict = {
        # version must be 1 at this time.
        # See https://docs.python.org/3/library/logging.config.html#configuration-dictionary-schema
        'version': 1,
        'disable_existing_loggers': False,
        'filters': {},
        'handlers': {},
        'loggers': {},
    }

    STREAM_HANDLER = {
        'level': 'DEBUG',
        'class': 'logging.StreamHandler',
        'formatter': 'Aries'
    }

    def __init__(self):
        self.config_dict = copy.deepcopy(self.default_config_dict)
        super().__init__()

    @staticmethod
    def __add_names(config, key, names):
        """Adds names to a logging config value
        The logging config value must be a list, e.g. handler names

        Args:
            config (dict): Logging config dict of a logger/handler, etc.
            key (str): The key in logging config containing a list of values.
            names (list): A list of strings to be added to the list of values.

        Example:
            config can be a logger config like:
            config = {
                'handlers': ['handler1'],
                'level': 'DEBUG',
                'propagate': True,
            }
            Calling __add_names(config, 'handlers', 'handler2') will add 'handler2'
            to the values of 'handlers' in the config. The method will return the following:
            {
                'handlers': ['handler1', 'handler2'],
                'level': 'DEBUG',
                'propagate': True,
            }
        """
        entries = config.get(key, [])
        for name in names:
            if name not in entries:
                entries.append(name)
        config[key] = entries
        return config

    def __add_entry(self, entry_key, entry_name, **kwargs):
        existing = self.config_dict.get(entry_key, dict())
        entry = {entry_name: kwargs}
        existing.update(entry)
        self.config_dict[entry_key] = existing
        return self

    def __update_entries(self, config_key, entry_key, names):
        entries = self.config_dict.get(config_key, {})
        self.config_dict[config_key] = {
            k: self.__add_names(v, entry_key, names) for k, v in entries.items()
        }
        return self

    def add_logger(self, logger_name, level="DEBUG", propagate=True, **kwargs):
        kwargs.update({
            "level": level,
            "propagate": propagate
        })
        self.__add_entry("loggers", logger_name, **kwargs)
        return self

    def add_handler(self, handler_name, handler_class, level="DEBUG", **kwargs):
        """Adds handler to all existing loggers
        
        Args:
            handler_name (str): Handler name
            handler_class (class): Handler class
            level:
        """
        kwargs.update({
            "class": handler_class,
            "level": level
        })
        self.__add_entry("handlers", handler_name, **kwargs)
        # Add handlers to existing loggers
        self.__update_entries("loggers", "handlers", [handler_name])
        return self

    def add_filters(self, filter_dict):
        """Adds filters to all existing handlers.
        If you do not want the filters to be added to some handlers,
        add the handlers after adding the filters.

        Args:
            filter_dict ([type]): [description]
        """
        # Add filters to existing filters
        existing_filters = self.config_dict.get("filters", dict())
        existing_filters.update(filter_dict)
        self.__update_entries("handlers", "filters", filter_dict.keys())
        return self

    def get_config(self):
        return self.config_dict



"""Contains tests for the excel module.
"""
import re
import logging
import string
import random
from shutil import copyfile

import os
import sys
aries_parent = os.path.join(os.path.dirname(__file__), "..", "..")
if aries_parent not in sys.path:
    sys.path.append(aries_parent)
from Aries.test import AriesTest
from Aries.excel import ExcelFile

logger = logging.getLogger(__name__)


class TestExcelFile(AriesTest):
    __test_file = os.path.join(os.path.dirname(__file__), "fixtures", "excel_test_file.xlsx")
    test_file = __test_file.replace("excel_test_file.xlsx", "excel_test.xlsx")

    def setUp(self):
        super().setUp()
        # Make a copy of the test file.
        # The original test file should never be used directly
        copyfile(self.__test_file, self.test_file)

    def tearDown(self):
        super().tearDown()
        if os.path.exists(self.test_file):
            os.remove(self.test_file)

    def assert_cell_value(self, file_path, cell, value):
        excel_file = ExcelFile(file_path)
        self.assertEqual(excel_file.worksheet[cell].value, value)


class TestReadExcelFile(TestExcelFile):

    def test_read_file_header(self):
        excel_file = ExcelFile(self.test_file)
        # Case sensitive
        self.assertTrue(excel_file.has_headers(["Header A", "Header AB", "header c"]))
        self.assertFalse(excel_file.has_headers(["header a", "header ab", "header c"]))
        # Case insensitive
        self.assertTrue(excel_file.has_headers(["header a", "header ab", "header c"], re.IGNORECASE))
        # Column index
        self.assertEqual(excel_file.column_index("header A"), 1)
        self.assertEqual(excel_file.column_index("Header AB"), 2)
        self.assertEqual(excel_file.column_index("Header A", case_sensitive=True), 1)
        self.assertEqual(excel_file.column_index("header A", case_sensitive=True), -1)
        # Use another row as index
        excel_file.set_headers(2)
        self.assertTrue(excel_file.has_headers(["row 1", "col 2"], re.IGNORECASE))

    def test_get_data_table(self):
        excel_file = ExcelFile(self.test_file)
        # Get data table with first row
        table = excel_file.get_data_table(skip_first_row=False)
        self.assertEqual(len(table), 3)
        self.assertEqual(len(table[0]), 3)
        # Get data table without first row
        table = excel_file.get_data_table()
        self.assertEqual(len(table), 2)


class TestWriteExcelFile(TestExcelFile):

    file_copy = TestExcelFile.test_file.replace("excel_test.xlsx", "excel_test_copy.xlsx")
    test_cell = 'C3'
    test_value = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8))

    def setUp(self):
        super().setUp()
        if os.path.exists(self.file_copy):
            os.remove(self.file_copy)

    def tearDown(self):
        super().tearDown()
        if os.path.exists(self.file_copy):
            os.remove(self.file_copy)

    def test_modify_file(self):
        copyfile(self.test_file, self.file_copy)
        excel_file = ExcelFile(self.file_copy)
        # Modify the file with a random value
        excel_file.worksheet[self.test_cell] = self.test_value
        excel_file.save()
        self.assert_cell_value(self.file_copy, self.test_cell, self.test_value)

    def test_save_as_another_file(self):
        excel_file = ExcelFile(self.test_file)
        excel_file.worksheet[self.test_cell] = self.test_value
        # Save a copy of the file
        excel_file.save(self.file_copy)
        self.assert_cell_value(self.file_copy, self.test_cell, self.test_value)

    def test_save_file_content(self):
        excel_file = ExcelFile(self.test_file)
        excel_file.worksheet[self.test_cell] = self.test_value
        # Save file content
        with open(self.file_copy, 'wb') as f:
            f.write(excel_file.content())
        self.assert_cell_value(self.file_copy, self.test_cell, self.test_value)

    def test_create_file(self):
        excel_file = ExcelFile()
        excel_file.worksheet[self.test_cell] = self.test_value
        # Save without a file path should raise an exception
        with self.assertRaises(Exception):
            excel_file.save()
        # Save with a file path
        excel_file.save(self.file_copy)
        self.assert_cell_value(self.file_copy, self.test_cell, self.test_value)




import logging
from ..web import WebAPI
logger = logging.getLogger(__name__)


class GoogleDriveFile:
    def __init__(self, access_token, file_id):
        self.file_id = file_id
        self._metadata = None
        self.api = WebAPI("https://www.googleapis.com/")
        self.api.add_header(Authorization="Bearer %s" % access_token)

    @property
    def metadata(self):
        """
        See Also: https://developers.google.com/drive/api/v3/fields-parameter
        """
        if not self._metadata:
            url = "https://www.googleapis.com/drive/v3/files/%s?fields=*" % self.file_id
            self._metadata = self.api.get_json(url)
        return self._metadata

    def get_meta(self, fields):
        """
        See Also: https://developers.google.com/drive/api/v3/fields-parameter
        """
        if isinstance(fields, list):
            fields = "&".join(["fields=%s" % f for f in fields])
            url = "https://www.googleapis.com/drive/v3/files/%s?%s" % (self.file_id, fields)
        else:
            url = "https://www.googleapis.com/drive/v3/files/%s?fields=%s" % (self.file_id, fields)
        return self.api.get_json(url)

    @property
    def revisions(self):
        return self.api.get_json("https://www.googleapis.com/drive/v3/files/%s/revisions" % self.file_id)


class GoogleSheet(GoogleDriveFile):
    @property
    def sheets(self):
        return self.get().get("sheets")

    def get(self, **kwargs):
        """Returns properties of the spreadsheet along with properties and merges of sheets in the spreadsheet.
        By default, data within grids will not be returned. You can include grid data one of two ways:
            1. Specify a field mask listing your desired fields using the fields parameter.
            2. Use key word argument "includeGridData=True".
        If a field mask is set, the includeGridData parameter is ignored
        For large spreadsheets, it is recommended to retrieve only the specific fields of the spreadsheet that you want.

        To retrieve only subsets of the spreadsheet, use the ranges parameter. Multiple ranges can be specified.
        Limiting the range will return only the portions of the spreadsheet that intersect the requested ranges.
        Ranges are specified using A1 notation.

        Args:
            **kwargs: ranges, includeGridData, fields

        Returns: A dictionary containing the information of the spreadsheet.

        See Also:
            https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/get
            https://developers.google.com/sheets/api/guides/concepts#partial_responses
        """
        api_url = "https://sheets.googleapis.com/v4/spreadsheets/%s" % self.file_id
        return self.api.get_json(api_url, **kwargs)

    def values(self, data_range, **kwargs):
        """Gets the values of a specific range.

        Args:
            data_range: A1 notation of a range in a sheet.
                e.g. "SheetName!A1:B5"

        Returns: A dictionary containing the following keys:
            majorDimension:
            range: The actual range in the sheet containing data.
            values: Two 2-D list containing the values.
        """
        return self.api.get_json(
            "https://sheets.googleapis.com/v4/spreadsheets/%s/values/%s" % (self.file_id, data_range),
            **kwargs
        )

    def get_data_grid(self, sheet_index=0, value_type="formattedValue"):
        sheets = self.get(includeGridData=True).get("sheets")
        if not sheets:
            return None

        data = sheets[sheet_index].get("data")
        if not data:
            return None

        rows = data[0].get("rowData", [])
        grid = []
        for row in rows:
            values = row.get("values", [])
            row_values = [v.get(value_type) for v in values]
            grid.append(row_values)
        return grid

    def get_column_data(self, sheet_name, col_idx, from_row=None):
        values = self.values("%s!%s:%s" % (sheet_name, col_idx, col_idx)).get("values")
        if not values:
            return []
        if from_row:
            values = values[from_row:]
        values = [v[0] if len(v) > 0 else "" for v in values]
        return values

    def append(self, data_range, rows):
        url = "https://sheets.googleapis.com/v4/spreadsheets/%s/values/%s:append?" \
              "valueInputOption=USER_ENTERED&insertDataOption=INSERT_ROWS" % (self.file_id, data_range)
        data = {
            "range": data_range,
            "majorDimension": "ROWS",
            "values": rows
        }
        self.api.post_json(url, data)



"""Provides unified IO and high level API for accessing folders and files.
"""
import os
import json
import logging
import binascii
import inspect
import traceback
from io import SEEK_SET, UnsupportedOperation
from io import BufferedIOBase, BufferedRandom, BufferedReader, BufferedWriter, TextIOWrapper
from .base import StorageObject, StorageFolderBase
from . import gs, file, web, s3
logger = logging.getLogger(__name__)


class StoragePrefix(StorageObject):
    """Represents a collections of object with the same URI prefix
    """
    # Maps the scheme to the underlying raw class.
    registry = {
        "file": file.LocalPrefix,
        "gs": gs.GSPrefix,
        "s3": s3.S3Prefix
    }

    def __init__(self, uri):
        """Initializes a StoragePrefix object with a URI prefix.

        Args:
            uri: URI prefix.
        """
        super(StoragePrefix, self).__init__(uri)
        raw_class = self.registry.get(self.scheme)
        if not raw_class:
            raise NotImplementedError("No implementation available for scheme: %s" % self.scheme)
        self.raw = raw_class(uri)

    @property
    def objects(self):
        """StorageFile objects with the same prefix, including files in sub-folders.
        """
        return self.raw.objects

    @property
    def files(self):
        """StorageFile objects having the same prefix but not under any sub-folder after the prefix
        """
        if hasattr(self.raw, "files"):
            return self.raw.files
        # Use objects to determine files if the files property is not supported by the raw class
        return [f for f in self.objects if "/" not in f.path.replace(self.prefix, "")]

    @property
    def folders(self):
        """StorageFolder objects having the same prefix but not under any sub-folder after the prefix
        """
        if hasattr(self.raw, "folders"):
            return self.raw.folders
        folder_paths = set()
        for f in self.objects:
            relative_path = f.path.replace(self.prefix, "")
            if "/" in relative_path:
                dir_name = relative_path.split("/", 1)[0]
                folder_paths.add(self.prefix + dir_name)
        return [StorageFolder(p) for p in folder_paths]

    @property
    def size(self):
        return self.raw.size

    def is_file(self):
        """Determine if the object is a file.
        This will return False if the object does not exist or the object is a folder.
        """
        if self.path.endswith("/"):
            return False
        if not self.exists():
            return False
        return True

    def exists(self):
        return self.raw.exists()

    def delete(self):
        return self.raw.delete()

    def copy(self, to):
        """Copies all the files by replacing the prefix
        """
        logger.debug("Copying files from %s to %s" % (self.uri, to))
        dest = StorageObject(to)
        if dest.scheme == self.scheme:
            try:
                return self.raw.copy(to)
            except (AttributeError, UnsupportedOperation):
                pass

        for storage_file in self.objects:
            storage_file.copy(storage_file.uri.replace(self.path, to))


class StorageFolder(StorageFolderBase):
    """Represents a storage folder.
    The StorageFolder class wraps an underlying raw class, which contains platform dependent implementation.
    The path attribute of a StorageFolder will always end with "/"

    """
    # Maps the scheme to the underlying raw class.
    registry = {
        "file": file.LocalFolder,
        "gs": gs.GSFolder,
        "s3": s3.S3Folder
    }

    def __init__(self, uri):
        """Initializes a StorageFolder.

        Args:
            uri: URI of the folder.
                The "file://" scheme will be used for local paths.
                A tailing slash "/" will be added to the path automatically.

        Examples:
            StorageFolder("gs://bucket_name/path/to/folder")
            StorageFolder("gs://bucket_name/path/to/folder/")
            StorageFolder("/path/to/local/folder")

        Raises:
            NotImplementedError: If there is no implementation for the scheme.

        """
        super(StorageFolder, self).__init__(uri)
        raw_class = self.registry.get(self.scheme)
        if not raw_class:
            raise NotImplementedError("No implementation available for scheme: %s" % self.scheme)
        self.raw = raw_class(uri)

    @staticmethod
    def init(uri):
        """Static method for initializing StorageFolder.
        """
        return StorageFolder(uri)

    @property
    def bucket_name(self):
        return self.__get_raw_attr("bucket_name", raise_error=True)

    @property
    def files(self):
        """

        Returns: A list of StorageFiles in the folder.

        """
        try:
            return self.raw.files
        except AttributeError:
            pass
        return [StorageFile(f) for f in self.file_paths]

    @property
    def folders(self):
        """

        Returns: A list of StorageFolders in the folder.

        """
        return [StorageFolder(f) for f in self.folder_paths]

    @property
    def objects(self):
        """Storage file objects in this folder and sub-folders.
        """
        try:
            return self.raw.objects
        except (AttributeError, UnsupportedOperation):
            pass

        object_list = self.files
        for folder in self.folders:
            object_list.extend(folder.objects)
        return object_list

    @property
    def size(self):
        try:
            return self.raw.size
        except (AttributeError, UnsupportedOperation):
            pass
        total = 0
        for f in self.files:
            s = f.size
            total += s if s else 0
        for folder in self.folders:
            s = folder.size
            total += s if s else 0
        return total

    @property
    def file_names(self):
        return [os.path.basename(f) for f in self.file_paths]

    @property
    def folder_names(self):
        return [str(f).rstrip("/").rsplit("/", 1)[-1] for f in self.folder_paths]

    @property
    def file_paths(self):
        return self.raw.file_paths

    @property
    def folder_paths(self):
        return self.raw.folder_paths

    def exists(self):
        """Checks if the folder exists.
        """
        return self.raw.exists()

    def create(self):
        """Creates a new folder.
        There will be no error if the folder already exists.
        """
        self.raw.create()
        return self

    def copy(self, to, contents_only=False):
        """Copies the folder.

        Args:
            to: The destination location,
            contents_only: Copies only the content of the folder.
                Defaults to False, i.e. a folder (with the same name as this folder)
                will be created at the destination to contain the files.

        Returns:

        """
        logger.debug("Copying files from %s to %s" % (self.uri, to))
        dest = StorageObject(to)
        if dest.scheme == self.scheme:
            try:
                return self.raw.copy(to, contents_only=contents_only)
            except (AttributeError, UnsupportedOperation):
                pass

        if not contents_only:
            to += self.name
        # logger.debug(self.file_paths)
        for storage_file in self.files:
            storage_file.copy(os.path.join(to, storage_file.basename))
        # logger.debug(self.folder_paths)
        # Recursively copy the sub-folders.
        for storage_folder in self.folders:
            storage_folder.copy(os.path.join(to, storage_folder.basename))

    def download(self, local_path):
        if not os.path.exists(local_path):
            os.makedirs(local_path)
        # Download the files using copy()
        for storage_file in self.files:
            storage_file.copy(os.path.join(local_path, storage_file.basename))
        # Recursively download the sub-folders.
        for storage_folder in self.folders:
            storage_folder.download(os.path.join(local_path, storage_folder.basename))

    def upload_from(self, local_path):
        raise NotImplementedError()

    def move(self, to, contents_only=False):
        """Moves the objects to another location."""
        self.copy(to, contents_only=contents_only)
        dest_folder = StorageFolder(to)
        if dest_folder.exists():
            # TODO: Check if the files are actually copied.
            if contents_only:
                self.empty()
            else:
                self.delete()
        else:
            raise FileNotFoundError("Failed to copy files to %s" % to)

    def delete(self):
        return self.raw.delete()

    def get_file_attributes(self, attribute=None):
        """Gets a list of files (represented by uri or other attribute) in the folder.

        Args:
            attribute: The attribute of the StorageFile to be returned in the list representing the files.

        Returns: A list of objects, each represents a file in this folder.
            If attribute is None, each object in the returning list will be the URI of the StorageFile
            If attribute is specified, each object in the returning list will be the attribute value of a StorageFile.

        """
        return self.get_attributes(self.files, attribute)

    def get_folder_attributes(self, attribute=None):
        """Gets a list of folders (represented by uri or other attribute) in the folder

        Args:
            attribute: The attribute of the StorageFolder to be returned in the list representing the folders.

        Returns: A list of objects, each represents a folder in this folder.
            If attribute is None, each object in the returning list will be the URI of the StorageFolder
            If attribute is specified, each object in the returning list will be the attribute value of a StorageFolder.

        """
        return self.get_attributes(self.folders, attribute)

    def get_folder(self, folder_name):
        """Gets a sub folder by name

        Args:
            folder_name (str): [description]

        Returns:
            LocalFolder: A LocalFolder instance of the sub folder.
                None if the sub folder does not exist.
        """
        for folder in self.folders:
            if folder.basename == folder_name:
                return folder
        return None

    def get_file(self, filename):
        for f in self.files:
            if f.basename == filename:
                return f
        return None

    def is_empty(self):
        if self.files or self.folders:
            return False
        else:
            return True

    def empty(self):
        for f in self.files:
            f.delete()
        for f in self.folders:
            f.delete()

    # Sub-class of StorageFolderBase can optionally implement the following methods
    #

    # def raw_if_exists(self, attr):
    #     from functools import wraps
    #
    #     @wraps(attr)
    #     def wrapper(*args, **kwargs):
    #         func_name = attr.__name__.rsplit(".", 1)[-1]
    #         if hasattr(self.raw, func_name):
    #             raw_attr = getattr(self.raw, func_name)(*args, **kwargs)
    #             if callable(raw_attr):
    #                 return raw_attr(*args, **kwargs)
    #             return raw_attr
    #         return attr(*args, **kwargs)
    #     return wrapper

    def __get_raw_attr(self, attr_name, raise_error=False):
        if hasattr(self.raw, attr_name):
            return getattr(self.raw, attr_name)
        if raise_error:
            raise AttributeError("%s:// does not support attribute %s" % (self.scheme, attr_name))
        return None

    def filter_files(self, prefix):
        attr_name = inspect.currentframe().f_code.co_name
        raw_attr = self.__get_raw_attr(attr_name)
        if raw_attr:
            return raw_attr(prefix)

        files = []
        for f in self.files:
            if f.name.startswith(prefix):
                files.append(f)
        logger.debug("%s files found with prefix %s" % (len(files), prefix))
        return files


class BufferedIOWrapper:
    """Delegates methods to buffered_io attribute.
    """
    def __init__(self):
        # Sub-class must define method to initialize buffered_io
        self.buffered_io = None

    def _check_closed(self):
        """Checks if the file is closed.

        Raises: ValueError if the file is closed.

        """
        if not self.buffered_io:
            raise ValueError(
                "I/O operation on closed file (buffer_io not initialized). "
                "Use open() or init() to open the file."
            )
        if self.buffered_io.closed:
            raise ValueError(
                "I/O operation on closed file (buffer_io closed). "
                "Use open() or init() to open the file."
            )

    def read1(self, size=None):
        self._check_closed()
        return self.buffered_io.read1(size)

    def readable(self):
        self._check_closed()
        return self.buffered_io.readable()

    def readinto(self, b):
        self._check_closed()
        return self.buffered_io.readinto(b)

    def readinto1(self, b):
        self._check_closed()
        return self.buffered_io.readinto1(b)

    def writable(self):
        self._check_closed()
        return self.buffered_io.writable()

    def write(self, b):
        self._check_closed()
        return self.buffered_io.write(b)

    def truncate(self, pos=None):
        self._check_closed()
        self.flush()
        if pos is None:
            pos = self.tell()
        return self.buffered_io.truncate(pos)

    def flush(self):
        self._check_closed()
        return self.buffered_io.flush()

    def detach(self):
        self._check_closed()
        return self.buffered_io.detach()

    def seekable(self):
        self._check_closed()
        return self.buffered_io.seekable()

    def seek(self, pos, whence=SEEK_SET):
        self._check_closed()
        return self.buffered_io.seek(pos, whence)

    def tell(self):
        self._check_closed()
        return self.buffered_io.tell()

    def fileno(self):
        self._check_closed()
        return self.buffered_io.fileno()

    def isatty(self):
        self._check_closed()
        return self.buffered_io.isatty()


class StorageFile(StorageObject, BufferedIOWrapper, BufferedIOBase):
    """Represents a storage file.

    Initializing a StorageFile object does not open the file.
    High level operations, e.g. copy(), delete() can be called without opening the file explicitly.
    read() is also supported without explicitly opening the file.
        If the file is closed, read() will open the file, read and return bytes, and then close the file.
        This is intended for one-time reading.

    To open a file for read/write, call open() explicitly.

    To initialize AND open a StorageFile, use StorageFile.init() static method.
    StorageFile.init() is designed be used in place of the python open()
    to obtain a file-like object for a file.
    See Also: https://docs.python.org/3/glossary.html#term-file-object

    To properly close the file, use context manager to with open() or init() when possible.

    Base on the scheme of the URI, StorageFile uses an sub-class of StorageIOBase as the underlying raw IO
    For binary mode, the raw IO is wrapped by BufferedIO
    For text mode, the bufferedIO is wrapped by TextIO

    Although StorageFile has raw_io as attribute, most operations should be done with the buffered_io

    Examples:
        Copy a file from local computer to Google storage:
        StorageFile("/path/to/file").copy("gs://path/to/file")

        Open a file and read all content:
        with StorageFile.init("/path/to/file") as f:
            content = f.read()

    Remarks:
        The order of the parent classes determines the method resolution order.
        See https://www.python.org/download/releases/2.3/mro/

    """

    registry = {
        "file": file.LocalFile,
        "gs": gs.GSFile,
        "s3": s3.S3File,
        "http": web.WebFile,
        "https": web.WebFile,
        "ftp": web.WebFile
    }

    def __init__(self, uri):
        """Initialize a StorageFile object.

        This will not open the file.
        However, high-level operations are supported without opening the file.

        Args:
            uri (str): The URI of the file, like file:///path/to/file or /path/to/file.
                If the URI does not contain a scheme, "file://" will be used.
        """
        StorageObject.__init__(self, uri)
        # Mode will be set by open()
        self._mode = None
        # The raw_io here is initialized but not opened.
        # The raw_io and buffered_io will be opened in __init_io()
        self.raw_io = self.__init_raw_io()
        # The buffered_io is the main IO stream used in the methods and properties.
        self.buffered_io = None

    def __call__(self, mode='r'):
        """Modifies the mode in which the file is open.
        The file will be closed and re-opened in the new mode.

        """
        if self.closed or not self._is_same_mode(mode):
            return self.open(mode)
        return self

    def __init_raw_io(self):
        """Initializes the underlying raw IO
        """
        # Create the underlying raw IO base on the scheme
        raw_class = self.registry.get(self.scheme)
        if not raw_class:
            raise NotImplementedError("No implementation available for scheme: %s" % self.scheme)
        return raw_class(self.uri)

    def __open_raw_io(self, closefd=True, opener=None):
        if not self.raw_io:
            self.__init_raw_io()

        # Raw IO always operates in binary mode, t and b will be ignored.
        mode = [c for c in self.mode if c in "rw+ax"]
        if self.scheme == "file":
            return self.raw_io.open(mode, closefd, opener)
        else:
            return self.raw_io.open(mode)

    def __buffer_size(self):
        """Determines the buffer size
        """

        buffering = self.BUFFER_SIZE
        # For local file only
        from .file import LocalFile
        if isinstance(self.raw_io, LocalFile):
            try:
                bs = os.fstat(self.raw_io.fileno()).st_blksize
            except (OSError, AttributeError):
                pass
            else:
                if bs > 1:
                    buffering = bs
        if buffering < 0:
            raise ValueError("Invalid buffering size: %s" % buffering)
        return buffering

    @staticmethod
    def __validate_args(text, binary, creating, reading, writing, appending, encoding, errors, newline, buffering):
        if text and binary:
            raise ValueError("Can't have text and binary mode at once")
        if creating + reading + writing + appending > 1:
            raise ValueError("Can't have read/write/append mode at once")
        if not (creating or reading or writing or appending):
            raise ValueError("Must have exactly one of read/write/append mode")
        if binary and encoding is not None:
            raise ValueError("Binary mode doesn't take an encoding argument")
        if binary and errors is not None:
            raise ValueError("Binary mode doesn't take an errors argument")
        if binary and newline is not None:
            raise ValueError("Binary mode doesn't take a newline argument")
        if binary and buffering == 1:
            import warnings
            warnings.warn("line buffering (buffering=1) isn't supported in binary "
                          "mode, the default buffer size will be used",
                          RuntimeWarning, 2)

    def __init_buffer_io(self, buffering, binary, updating, creating, reading, writing, appending):
        if buffering < 0:
            buffering = self.__buffer_size()
        if buffering == 0:
            if binary:
                # logger.debug("Using raw_io as buffer_io for %s ..." % self.uri)
                return self.raw_io
            raise ValueError("can't have unbuffered text I/O")
        if updating:
            buffered_io = BufferedRandom(self.raw_io, buffering)
        elif creating or writing or appending:
            buffered_io = BufferedWriter(self.raw_io, buffering)
        elif reading:
            buffered_io = BufferedReader(self.raw_io, buffering)
        else:
            raise ValueError("Unknown mode: %r" % self.mode)
        # logger.debug("Initialized buffer IO for %s" % self.uri)
        return buffered_io

    def __init_io(self, buffering=-1, encoding=None, errors=None, newline=None, closefd=True, opener=None):
        """Initializes the underlying raw IO and buffered IO
        """
        # The following code is modified based on python io.open()
        modes = set(self.mode)
        if modes - set("axrwb+t") or len(self.mode) > len(modes):
            raise ValueError("invalid mode: %r" % self.mode)

        creating = "x" in modes
        reading = "r" in modes
        writing = "w" in modes
        appending = "a" in modes
        updating = "+" in modes
        text = "t" in modes
        binary = "b" in modes

        self.__validate_args(text, binary, creating, reading, writing, appending, encoding, errors, newline, buffering)
        self.raw_io = self.__open_raw_io(closefd, opener)

        # Track the opened IO
        opened_io = self.raw_io
        try:
            line_buffering = False
            if buffering == 1 or buffering < 0 and self.raw_io.isatty():
                buffering = -1
                line_buffering = True
            opened_io = self.__init_buffer_io(buffering, binary, updating, creating, reading, writing, appending)

            # Use TextIOWrapper for text mode
            if binary:
                return opened_io
            text_io = TextIOWrapper(opened_io, encoding, errors, newline, line_buffering)
            opened_io = text_io
            text_io.mode = self.mode
            return opened_io
        except Exception as ex:
            # Close the opened IO if there is an error
            if opened_io:
                logger.debug("Closing opened_io...")
                opened_io.close()
            raise ex

    def _is_same_mode(self, mode):
        """Checks if the mode is the same as the one in which the file is open
        """
        if not self.mode:
            return False
        if mode:
            return sorted(self.mode) == sorted(mode)
        return True

    def open(self, mode='r', buffering=-1, encoding=None, errors=None, newline=None, closefd=True, opener=None):
        """Opens a file and initializes the underlying IO.

        Args:
            mode (str): The mode in which the file is opened,
                must be the combination of r, w, x, a, b, t and +
            buffering (int): Buffering policy.
            encoding (str): Name of the encoding.
            errors (str): The way to handle encoding error, must be:
                'strict': (Default) Raise a ValueError exception
                'ignore': Ignore errors.
            newline (str): New line character, which can be '', '\n', '\r', and '\r\n'.
            closefd (bool): This option currently has no effect.
                The underlying file descriptor will always be closed.
            opener: Custom opener.

        Returns: self (StorageFile)

        """
        # Stores the arguments as protected attributes
        self._mode = str(mode)
        # logger.debug("Opening %s ..." % self.uri)
        # The raw_io will be initialized in __init_io()
        if not self.raw_io.closed:
            self.raw_io.close()

        # The buffered_io is the main IO stream used in the methods and properties.
        self.buffered_io = self.__init_io(buffering, encoding, errors, newline, closefd, opener)
        return self

    @staticmethod
    def init(uri, mode='r', **kwargs):
        """Initializes the StorageFile and open the underlying IO.
        This is a simplified version of StorageFile(uri).open(mode)

        Args:
            uri (str): The URI of the file, like file:///path/to/file or /path/to/file.
                If the URI does not contain a scheme, "file://" will be used.
            mode (str): The mode in which the file is opened,
                must be the combination of r, w, x, a, b, t and +

        Returns: A StorageFile object

        """
        storage_file = StorageFile(uri)
        storage_file.open(mode, **kwargs)
        return storage_file

    @staticmethod
    def load_json(uri):
        """Loads a json file into a dictionary.
        """
        return json.loads(StorageFile.init(uri).read())

    @property
    def closed(self):
        if not self.buffered_io:
            return True
        return self.buffered_io.closed

    @property
    def raw(self):
        """The underlying Raw IO
        """
        return self.raw_io

    @property
    def mode(self):
        return self._mode

    @property
    def size(self):
        """Size of the file in bytes.
        None will be returned if the size cannot be determined.
        """
        if hasattr(self.raw_io, "size"):
            try:
                return self.raw_io.size
            except Exception as ex:
                traceback.print_exc()
                logger.debug("Failed to get size: %s" % ex)
                pass
        return None

    @property
    def md5_hex(self):
        """MD5 hex of the file content
        None will be returned if MD5 is not available.
        """
        return self.raw.md5_hex

    @property
    def blob(self):
        """The blob of the raw IO.

        Raises: AttributeError if blob is not supported.

        """
        if hasattr(self.raw_io, "blob"):
            return self.raw_io.blob
        raise AttributeError("%s:// does not support blob attribute" % self.scheme)

    @property
    def bucket_name(self):
        """Bucket name of the raw IO

        Raises: AttributeError if blob is not supported.

        """
        if hasattr(self.raw_io, "bucket_name"):
            return self.raw_io.bucket_name
        raise AttributeError("%s:// does not support bucket_name attribute" % self.scheme)

    @property
    def updated_time(self):
        return self.raw_io.updated_time

    def exists(self):
        """Checks if the file exists.
        This method may not return True until a new file is closed.
        """
        # TODO: File may not exist until new file is closed.
        return self.raw_io.exists()

    def __enter__(self):
        self._check_closed()
        return self

    def __exit__(self, exc_type, exc_value, trace):
        self.close()
        return

    def create(self):
        return self.raw_io.create()

    def delete(self):
        """Deletes the file.

        Returns:

        """
        if self.buffered_io:
            self.buffered_io.close()
        self.raw_io.delete()

    def copy(self, to):
        logger.debug("Copying file to %s" % to)
        dest_file = StorageFile(to)
        # Use raw_io copy for same scheme, if possible
        if self.scheme == dest_file.scheme and hasattr(self.raw_io, "copy"):
            return self.raw_io.copy(to)
        with self.open("rb") as f:
            with dest_file.open('wb') as f_to:
                return self.copy_stream(f, f_to)

    def move(self, to):
        """Moves the objects to another location."""
        self.copy(to)
        dest_file = StorageFile(to)
        if dest_file.exists():
            self.delete()
        else:
            raise FileNotFoundError("Failed to copy files to %s" % to)

    def local(self):
        """Creates a temporary local copy of the file to improve the performance.
        This requires the supports from the underlying raw IO.
        """
        try:
            self.raw_io.local()
        except (AttributeError, UnsupportedOperation):
            pass
        return self

    def download(self, to_file_obj=None, **kwargs):
        """Downloads the file to a file-object.

        Args:
            to_file_obj: A file-object. A NamedTemporaryFile will be created if this is None

        Returns: The file-object.

        """
        if not self.raw_io.exists():
            raise FileNotFoundError("File %s not found." % self.uri)

        # Create a new temp file if to_file_obj is None
        if not to_file_obj:
            to_file_obj = self.create_temp_file(**kwargs)
        try:
            logger.debug("Downloading %s ..." % self.uri)
            self.raw_io.download(to_file_obj)
            to_file_obj.flush()
            return to_file_obj
        except (AttributeError, UnsupportedOperation):
            pass
        # Copy the stream
        with self.open("rb") as f:
            self.copy_stream(f, to_file_obj)
        return to_file_obj

    def upload(self, from_file_obj):
        try:
            return self.raw_io.upload(from_file_obj)
        except (AttributeError, UnsupportedOperation):
            pass
        if not self.closed:
            self.buffered_io.close()
        with self.open("wb") as f:
            f.raw_io.load_from(from_file_obj)

    def upload_from_file(self, file_path):
        with open(file_path, 'rb') as f:
            self.upload(f)

    def is_gz(self):
        """Determine if the file is gz compressed.
        """
        # Reset the offset to the beginning of the file if file is opened.
        if self.closed:
            with self.open("rb") as f:
                b = f.read(2)
        else:
            offset = self.tell()
            self.seek(0)
            b = self.read(2)
            # Move offset back
            self.seek(offset)
        b = binascii.hexlify(b)
        logger.debug("File begins with: %s" % b)
        return b == b'1f8b'

    def close(self):
        # logger.debug("Closing %s ..." % self.uri)
        results = None
        if self.buffered_io:
            # buffered_io will close raw_io
            results = self.buffered_io.close()
            # Remove the buffered_io reference so that it will close the raw IO
            self.buffered_io = None
        elif self.raw_io and not self.raw_io.closed:
            self.raw_io.close()
        return results

    def read(self, size=None):
        # As a shortcut, read() can be called without initializing buffered_io
        if self.closed:
            # The returned content will always be bytes in this case.
            with self.raw_io.open('rb') as f:
                return f.read(size)
        return self.buffered_io.read(size)
